#
# LNXHC::Util.pm
#   Linux Health Checker utility functions for use by framework and plug-ins
#
# Copyright IBM Corp. 2012
#
# Author(s): Peter Oberparleiter <peter.oberparleiter@de.ibm.com>
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#
# Contributors: See file CONTRIBUTORS which is part of this package
#

package LNXHC::Util;

use strict;
use warnings;

use Exporter qw(import);


#
# Local imports
#


#
# Default exports
#
our @EXPORT;


#
# On-demand exports
#
our @EXPORT_OK = qw($ALIGN_T_CENTER $ALIGN_T_LEFT $ALIGN_T_RIGHT $FIELD_T_ALIGN
		    $FIELD_T_DELIM $FIELD_T_MAX $FIELD_T_MIN $FIELD_T_WEIGHT
		    $LAYOUT_T_FIELDS $LAYOUT_T_FMT $LAYOUT_T_WIDTH
		    $LAYOUT_T_WIDTHS &layout_get_width &lprintf &lsprintf);

our %EXPORT_TAGS = (
	consumer => [qw($ALIGN_T_CENTER $ALIGN_T_LEFT $ALIGN_T_RIGHT
			$FIELD_T_ALIGN $FIELD_T_DELIM $FIELD_T_MAX $FIELD_T_MIN
			$FIELD_T_WEIGHT $LAYOUT_T_FIELDS $LAYOUT_T_FMT
			$LAYOUT_T_WIDTH $LAYOUT_T_WIDTHS
			&layout_get_width &lprintf &lsprintf)],
);

#
# Constants
#

# Define alignment for layout fields
# Enumeration (enum align_t)

# Item is aligned to the left
our $ALIGN_T_LEFT		= 0;
# Item is aligned to the right
our $ALIGN_T_RIGHT		= 1;
# Item is centered
our $ALIGN_T_CENTER		= 2;


# Define layout for a single field
# Enumeration of data fields for struct field_t

# Minimum number of characters required by this field
our $FIELD_T_MIN		= 0;	# unsigned int
# Maximum number of characters required by this field or undef for unlimited
our $FIELD_T_MAX		= 1;	# unsigned int
# Weight determining share of extra space assigned to this field
our $FIELD_T_WEIGHT		= 2;	# unsigned int
# Item alignment
our $FIELD_T_ALIGN		= 3;	# enum align_t
# Right-hand delimiter string for this field
our $FIELD_T_DELIM		= 4;	# string


# Layout definition
# Enumeration of data fields for struct layout_t

# List of field layouts
our $LAYOUT_T_FIELDS		= 0;	# struct field_t[]
# List of resulting field widths
our $LAYOUT_T_WIDTHS		= 1;	# unsigned int[]
# Format string representing fields
our $LAYOUT_T_FMT		= 2;	# string
# Total width
our $LAYOUT_T_WIDTH		= 3;	# unsigned int


#
# Global variables
#

my $_columns = $ENV{"COLUMNS"};


#
# Sub-routines
#

#
# _layout_prepare(layout)
#
# Prepare LAYOUT for use by layout_printf.
#
sub _layout_prepare($)
{
	my ($layout) = @_;
	my $fields = $layout->[$LAYOUT_T_FIELDS];
	my $field;
	my @widths;
	my $fmt;
	my @todos;
	my $free = $_columns;
	my $total_weight;
	my $total_width;
	my $i;

	# Initialize widths array
	$i = 0;
	$total_weight = 0;
	foreach $field (@$fields) {
		my ($min, $max, $weight, $align, $delim) = @$field;

		push(@widths, $min);
		if ($weight > 0 && (!defined($max) || $max != $min)) {
			# Variable width field, either min..max or
			# min..max share of available space
			push(@todos, $i);
			$total_weight += $weight;
		}
		$free -= $min + length($delim);
		$i++;
	}

	# Distribute free space until either all fields are satisfied or
	# no free space is left
	while (@todos && $free > 0) {
		my @new_todos;
		my $new_total_weight = 0;
		my $new_free = $free;
		my %fract_db;

		# Distribute free space among fields which still have place
		foreach $i (@todos) {
			my $max = $fields->[$i]->[$FIELD_T_MAX];
			my $weight = $fields->[$i]->[$FIELD_T_WEIGHT];
			my $width = $widths[$i];
			my $share = ($free * $weight) / $total_weight;
			my $fract = $share - int($share);

			$share = int($share);
			if (defined($max) && $width + $share > $max) {
				# Field is full
				$share = $max - $width;
			}
			$width += $share;
			$new_free -= $share;

			if (!defined($max) || $width < $max) {
				# There's more room
				push(@new_todos, $i);
				$new_total_weight += $weight;

				# Fract will be used to determine order when
				# distributing rounding error leftovers
				$fract_db{$i} = $fract;
			}
			$widths[$i] = $width;
		}

		# In case no field gained any share, distribute leftovers
		# according to share fractions to prevent an endless loop
		if ($free == $new_free) {
			while (%fract_db && $new_free > 0) {
				foreach $i (sort( {
					  $fract_db{$a} == $fract_db{$b} ?
						$a <=> $b :
						$fract_db{$b} <=> $fract_db{$a}
					  } keys(%fract_db))) {
					my $max = $fields->[$i]->
							[$FIELD_T_MAX];
					$widths[$i]++;
					if (defined($max) &&
					    $widths[$i] == $max) {
						delete($fract_db{$i});
					}
					$new_free--;
					if ($new_free == 0) {
						last;
					}
				}
			}
			last;
		}

		@todos		= @new_todos;
		$total_weight	= $new_total_weight;
		$free		= $new_free;
	}

	# Build format string
	$fmt = "";
	$total_width = 0;
	for ($i = 0; $i < scalar(@$fields); $i++) {
		my $field = $fields->[$i];
		my $align = $field->[$FIELD_T_ALIGN];
		my $delim = $field->[$FIELD_T_DELIM];
		my $width = $widths[$i];

		if ($align == $ALIGN_T_RIGHT) {
			$fmt .= "%".$width."s".$delim;
		} else {
			$fmt .= "%-".$width."s".$delim;
		}
		$total_width += $width + length($delim);
	}

	# Store results
	$layout->[$LAYOUT_T_WIDTHS]	= \@widths;
	$layout->[$LAYOUT_T_WIDTH]	= $total_width;
	$layout->[$LAYOUT_T_FMT]	= $fmt;
}

#
# lsprintf(layout, @args)
#
# Return string representation of ARGS formatted according to LAYOUT.
#
sub lsprintf($@)
{
	my ($layout, @args) = @_;
	my ($fields, $widths, $fmt) = @$layout;
	my $arg;
	my $i;

	if (!defined($widths)) {
		# Do this only once
		_layout_prepare($layout);
		($fields, $widths, $fmt) = @$layout;
	}

	$i = 0;
	foreach $arg (@args) {
		if (length($arg) > $widths->[$i]) {
			$arg = substr($arg, 0, $widths->[$i] - 3)."...";
		} elsif ($fields->[$i]->[$FIELD_T_ALIGN] == $ALIGN_T_CENTER) {
			my $pad = ($widths->[$i] - length($arg)) / 2;

			$arg = (" "x$pad).$arg;
		}
		$i++;
	}

	# Fill in missing args
	while ($i < scalar(@$widths)) {
		push(@args, "");
		$i++;
	}

	return sprintf($fmt, @args);
}

#
# lprintf(layout, @args)
#
# Print string representation of ARGS formatted according to LAYOUT.
#
sub lprintf($@)
{
	my ($layout, @args) = @_;

	print(lsprintf($layout, @args));
}

#
# layout_get_width(layout)
#
# Return total width of LAYOUT.
#
sub layout_get_width($)
{
	my ($layout) = @_;
	my $width = $layout->[$LAYOUT_T_WIDTH];

	if (!defined($width)) {
		# Do this only once
		_layout_prepare($layout);
		$width = $layout->[$LAYOUT_T_WIDTH];
	}

	return $width;
}


#
# Code entry
#

# Indicate successful module initialization
1;
